import os
import re
import shutil
from collections.abc import ItemsView, KeysView, ValuesView
from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Generator, Union

import tomlkit
from astropy.coordinates import EarthLocation
from astropy.units import Quantity
from tomlkit.items import Table
from tomlkit.toml_document import TOMLDocument

from . import EnvVarName, get_logger
from .exceptions import NECSTConfigurationError
from .utils import ValueRange, read_file

logger = get_logger(__name__)

DefaultNECSTRoot = Path.home() / ".necst"


class _Parsers:
    observatory = str
    location = lambda x: EarthLocation(**x)  # noqa: E731
    observation_frequency = Quantity
    simulator = bool
    alert_interval_sec = float
    antenna_pid_param_az = list
    antenna_pid_param_el = list
    antenna_drive_range_az = lambda x: ValueRange(*map(Quantity, x))  # noqa: E731
    antenna_drive_range_el = lambda x: ValueRange(*map(Quantity, x))  # noqa: E731
    antenna_drive_warning_limit_az = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_drive_warning_limit_el = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_drive_critical_limit_az = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_drive_critical_limit_el = lambda x: ValueRange(  # noqa: E731
        *map(Quantity, x)
    )
    antenna_pointing_accuracy = Quantity
    antenna_pointing_parameter_path = Path
    antenna_scan_margin = Quantity
    antenna_max_acceleration_az = Quantity
    antenna_max_acceleration_el = Quantity
    antenna_max_speed_az = Quantity
    antenna_max_speed_el = Quantity
    antenna_speed_to_pulse_factor_az = Quantity
    antenna_speed_to_pulse_factor_el = Quantity
    antenna_command_frequency = int
    antenna_command_offset_sec = float
    ros_service_timeout_sec = float
    ros_communication_deadline_sec = float
    ros_logging_interval_sec = float
    ros_topic_scan_interval_sec = float


class Configuration:

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self.__config = None
        self.reload()

    def __getitem__(self, key: str) -> Any:
        ret = self.__config.__getitem__(key)
        return self._dotnecst / ret if isinstance(ret, Path) else ret

    def __getattr__(self, key: str) -> Any:
        ret = getattr(self.__config, key)
        return self._dotnecst / ret if isinstance(ret, Path) else ret

    @property
    def _dotnecst(self) -> Path:
        return self.__config_path.parent

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(file='{self.__config_path}')"

    def __str__(self) -> str:
        return str(self.__config)

    def reload(self) -> None:
        self.__config_path = self.__find_config()
        params = tomlkit.parse(read_file(self.__config_path, localonly=True))
        new_config = _Cfg(self, params)
        # Attribute validation
        attr_cross_section = set(dir(self.__class__)) & set(new_config.keys())
        dupl_attrs = list(filter(lambda x: not x.startswith("_"), attr_cross_section))
        if dupl_attrs:
            raise NECSTConfigurationError(
                f"Parameters {dupl_attrs} cannot be set; reserved name."
            )
        # Assign after validation
        self.__config = new_config

    def __find_config(self) -> Path:
        root_candidates = [DefaultNECSTRoot]
        if EnvVarName.necst_root in os.environ:
            root_candidates.insert(0, os.environ[EnvVarName.necst_root])
        config_file_name = os.environ.get(EnvVarName.necst_config_name, "config.toml")
        candidates = map(lambda x: str(x) + f"/{config_file_name}", root_candidates)

        for path in candidates:
            try:
                read_file(path, localonly=True)
                found = Path(re.sub(r".*://", "", path))
                logger.info(f"Imported configuration file '{found}'")
                return found
            except FileNotFoundError:
                pass

            try:
                saveto = DefaultNECSTRoot / "config.toml"
                read_file(path, localonly=False, saveto=saveto, overwrite=True)
                logger.info(
                    f"Imported configuration file '{path}', "
                    f"which was saved to '{saveto}'"
                )
                return saveto
            except FileNotFoundError:
                pass

        logger.error(
            "Config file not found, using the default parameters. "
            "To create the file with default parameters, run `neclib.configure()`."
        )
        return Path(__file__).parent / "src" / "config.toml"

    @classmethod
    def configure(cls) -> None:
        """Create config file under ``$HOME/.necst``"""
        DefaultNECSTRoot.mkdir(exist_ok=True)
        for filename in ["config.toml", "pointing_param.toml"]:
            _target_path = DefaultNECSTRoot / filename
            if _target_path.exists():
                logger.error(f"'{_target_path}' already exists, skipping...")
                continue
            shutil.copyfile(Path(__file__).parent / "src" / filename, _target_path)
        cls().reload()


class TOMLDict:
    def __init__(self, data: TOMLDocument) -> None:
        self._toml_data = data
        self._native_data = self._toml_data.unwrap()
        for k, v in self.flat:
            *namespace, _k = k.split(".")
            ref = self._native_data
            for ns in namespace:
                ref = ref[ns]
            v = v.unwrap() if hasattr(v, "unwrap") else v
            ref[_k] = self.get_parser(k)(v)

    def get_parser(self, key: str) -> Callable[[Any], Any]:
        return getattr(_Parsers, key.replace(".", "_"), lambda x: x)

    def get(self, key: str, full: bool = False, parse: bool = False) -> Any:
        key = key.lower()

        *namespace, key = key.split(".")
        extracted = deepcopy(self._native_data) if parse else deepcopy(self._toml_data)
        ref = extracted
        for ns in namespace:
            for k in ref.copy():
                if ns != k.lower():
                    ref.pop(k)
            ref = ref[ns]

        if key in ref:
            for k in ref.copy():
                if key != k.lower():
                    ref.pop(k)
            return extracted if full else ref[key]
        else:
            for k in ref.copy():
                if not k.lower().startswith(key):
                    ref.pop(k)
            return extracted if full else ref

    def set(self, key: str, value: Any) -> None:
        *namespace, key = key.split(".")
        ref_toml = self._toml_data
        ref_native = self._native_data
        for ns in namespace:
            if ns not in ref_toml:
                ref_toml[ns] = Table()
                ref_native[ns] = {}
            ref_toml = ref_toml[ns]
            ref_native = ref_native[ns]
        ref_toml[key] = value
        ref_native[key] = self.get_parser(key)(value)

    def __setitem__(self, key: str, value: Any) -> None:
        self.set(key, value)

    @property
    def flat(self) -> Generator[Any, None, None]:
        def generate_flat(data: Any, prefix: str = "") -> Generator[Any, None, None]:
            for key, value in data.items():
                if isinstance(value, Table):
                    yield from generate_flat(value, prefix + key + ".")
                else:
                    yield prefix + key, self._native_data.get(prefix + key, value)

        yield from generate_flat(self._toml_data)

    @property
    def raw_flat(self) -> Generator[Any, None, None]:
        def generate_flat(data: Any, prefix: str = "") -> Generator[Any, None, None]:
            for key, value in data.items():
                if isinstance(value, Table):
                    yield from generate_flat(value, prefix + key + ".")
                else:
                    yield prefix + key, value

        yield from generate_flat(self._toml_data)

    def flat_keys(self) -> Generator[str, None, None]:
        for key, _ in self.flat:
            yield key

    raw_flat_keys = flat_keys

    def flat_values(self) -> Generator[Any, None, None]:
        for _, value in self.flat:
            yield value

    def raw_flat_values(self) -> Generator[Any, None, None]:
        for _, value in self.raw_flat:
            yield value

    flat_items = flat

    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    def __str__(self) -> str:
        return str(self._toml_data)

    def __repr__(self) -> str:
        return repr(self._toml_data)

    @property
    def raw(self) -> TOMLDocument:
        return self._toml_data


class _Cfg:
    __slots__ = ["_config", "_prefix", "_config_manager"]

    def __init__(
        self,
        config_manager: Configuration,
        config: Union[Table, TOMLDocument],
        prefix: str = "",
    ) -> None:
        self._prefix = prefix
        self._config = TOMLDict(config)
        self._config_manager = config_manager

        config_keys = [key.lower() for key in self._config.flat_keys()]
        duplicated_key = [key for key in set(config_keys) if config_keys.count(key) > 1]
        if len(duplicated_key) > 0:
            raise NECSTConfigurationError(
                "Duplicated keys are not allowed in the configuration file: "
                f"{duplicated_key}"
            )

    def __getattr__(self, key: str) -> Any:
        return self.__getitem__(key)

    @property
    def _dotnecst(self) -> Path:
        return self._config_manager._dotnecst

    def __getitem__(self, key: str) -> Any:
        full_key = (self._prefix + key).lower()
        config_keys = [k.lower() for k in self._config.flat_keys()]
        if full_key in config_keys:
            ret = self._config.get(full_key, False, True)
            return self._dotnecst / ret if isinstance(ret, Path) else ret
        key_is_namespace = bool(
            [k for k in config_keys if k.startswith(full_key + ".")]
        )
        if key_is_namespace:
            return _Cfg(
                self._config_manager, self._config.get(full_key, True), full_key + "."
            )
        existent_key = [k for k in config_keys if k.startswith(full_key)]
        if existent_key:
            return _Cfg(
                self._config_manager, self._config.get(full_key, True), full_key + "_"
            )
        return None

    def __str__(self) -> str:
        width = max(len(p) for p in self._config.flat_keys()) - len(self._prefix) + 2

        def _prretify(key: str) -> str:
            key, value = key[len(self._prefix) :], self._config.get(key, False, True)
            return f"    {key:{width}s}{value!s}    ({type(value).__name__})"

        _parameters = "\n".join(map(_prretify, self._config.flat_keys()))
        return f"NECST configuration (prefix='{self._prefix}')\n{_parameters}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(prefix='{self._prefix}')"

    def keys(self, full: bool = False) -> KeysView:
        flattened = self._config.flat
        if not full:
            prefix_len = len(self._prefix)
            flattened = map(lambda x: (x[0][prefix_len:], x[1]), flattened)
        return dict(flattened).keys()

    def values(self, raw: bool = False) -> ValuesView:
        flattened = self._config.raw_flat if raw else self._config.flat
        flattened = map(
            lambda x: (x[0], self._dotnecst / x[1] if isinstance(x[1], Path) else x[1]),
            flattened,
        )
        if raw:
            return dict(flattened).values()
        return dict((k, self[k]) for k, _ in flattened).values()

    def items(self, full: bool = False, raw: bool = False) -> ItemsView:
        flattened = self._config.raw_flat if raw else self._config.flat
        flattened = map(
            lambda x: (x[0], self._dotnecst / x[1] if isinstance(x[1], Path) else x[1]),
            flattened,
        )
        if not full:
            prefix_len = len(self._prefix)
            flattened = map(lambda x: (x[0][prefix_len:], x[1]), flattened)
        if raw:
            return dict(flattened).items()
        return dict((k, self[k]) for k, _ in flattened).items()

    def __gt__(self, other: Any) -> bool:
        if other is None:
            return True
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented
        if set(self.keys(full=False)) <= set(other.keys(full=False)):
            return False
        lazy_eq = (getattr(self, k) == getattr(other, k) for k in other.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in other.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __lt__(self, other: Any) -> bool:
        if other is None:
            return False
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented
        if set(self.keys(full=False)) >= set(other.keys(full=False)):
            return False
        lazy_eq = (getattr(self, k) == getattr(other, k) for k in self.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in self.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented
        if set(self.keys(full=False)) != set(other.keys(full=False)):
            return False
        lazy_eq = (getattr(self, k) == getattr(other, k) for k in self.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in self.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __ge__(self, other: Any) -> bool:
        return self.__eq__(other) or self.__gt__(other)

    def __le__(self, other: Any) -> bool:
        return self.__eq__(other) or self.__lt__(other)

    def __add__(self, other: Any) -> "_Cfg":
        if other is None:
            return self
        if not isinstance(other, (_Cfg, Configuration)):
            return NotImplemented
        if self._prefix == other._prefix:
            new_config = TOMLDict(deepcopy(self._config))
            for k, v in other.items(full=True, raw=True):
                new_config[k] = v
            for k, v in self.items(full=True, raw=True):
                if (k in new_config.flat_keys()) and (new_config.get(k, False) != v):
                    raise ValueError(
                        "Cannot merge configurations with different values"
                    )
                new_config[k] = v
            return _Cfg(self._config_manager, new_config.raw, self._prefix)
        else:
            new_config = TOMLDict(TOMLDocument())
            for k, v in self.items(full=False, raw=True):
                new_config[k] = v
            for k, v in other.items(full=False, raw=True):
                if (k in new_config.flat_keys()) and (new_config.get(k, False) != v):
                    raise ValueError(
                        "Cannot merge configurations with different values"
                    )
                new_config[k] = v
            return _Cfg(self._config_manager, new_config.raw, "")

    __radd__ = __add__


config = Configuration()
configure = Configuration.configure



from typing import Callable, Union

from ...utils import busy, sanity_check
from .da_converter_base import DAConverter


class CPZ340816(DAConverter):

    Manufacturer = "Interface"
    Model = "CPZ340816"

    Identifier = "rsw_id"

    def __init__(self, **kwargs):
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.param_buff = {i: 0.0 for i in range(1, 17)}  # All in [V]
        self.da = pyinterface.open(3408, self.rsw_id)

    @property
    def converter(self) -> Callable[[Union[int, float]], float]:
        sanity_check(self.Config.converter, "x")
        return eval(f"lambda x: {self.Config.converter}")

    def set_voltage(self, mV: float, id: str) -> None:
        ch = self.Config.channel[id]
        if ch not in self.param_buff.keys():
            raise ValueError(f"Invaild channel {ch}")
        if not self.Config.max_mv[0] < mV < self.Config.max_mv[1]:
            raise ValueError(f"Unsafe voltage {mV} mV")
        else:
            self.param_buff[ch] = self.converter(mV)

    def apply_voltage(self) -> None:
        with busy(self, "busy"):
            for i in range(0, 16):
                ch = int(list(self.param_buff.keys())[i])
                voltage = list(self.param_buff.values())[i]

                self.da.output_da(ch, voltage)

    def finalize(self) -> None:
        self.da.finalize()
        self.param_buff = {i: 0.0 for i in range(1, 17)}



import time

import astropy.units as u
import ogameasure

from ...utils import busy
from .thermometer_base import Thermometer


class Model218(Thermometer):

    Manufacturer = "LakeShore"
    Model = "Model218"

    Identifier = "host"

    def __init__(self, **kwargs) -> None:
        self.thermometer = ogameasure.Lakeshore.model218_usb(self.Config.usb_port)

    def get_temp(self, id: str) -> u.Quantity:
        ch = self.Config.channel[id]
        with busy(self, "busy"):
            data = self.thermometer.kelvin_reading_query(ch)
            time.sleep(0.1)
            return data * u.K

    def finalize(self) -> None:
        self.thermometer.com.close()



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class Attenuator(DeviceBase):
    @abstractmethod
    def set_loss(self, dB: int, id: str):
        ...

    @abstractmethod
    def get_loss(self, id: str) -> u.Quantity:
        ...



from abc import ABC, abstractmethod
from collections import UserDict
from functools import partial
from typing import Any, ClassVar, Dict, List, Optional, Union, final

from ..configuration import Configuration, _Cfg, config


def get_device_list() -> Dict[str, "DeviceBase"]:
    # Unlike __mro__, __subclasses__ returns immediate subclasses only.
    device_kinds = DeviceBase.__subclasses__()
    table = {}
    for v in device_kinds:
        for impl in v.__subclasses__():
            table[impl.Model.lower()] = impl
    return table


def find_config(
    key: str, identifier: Optional[str] = None
) -> Union[Configuration, _Cfg]:
    if identifier is None:
        return config[key]
    devices = {k[:-2]: v for k, v in config.items() if k.endswith("._")}
    same_model = {k: config[k] for k in devices if config[k]._ == config[key]._}
    same_machine = {
        k: v
        for k, v in same_model.items()
        if config[k][identifier] == config[key][identifier]
    }
    cfg = None
    for _cfg in same_machine.values():
        cfg += _cfg
    return cfg


class DeviceMapping(UserDict):
    def __getattr__(self, key: str) -> Dict[str, Any]:
        targets = {k: getattr(v, key) for k, v in self.items()}
        if all(callable(t) for t in targets.values()):

            def func(*args, **kwargs) -> List[Any]:
                if (
                    ("id" in kwargs)
                    and isinstance(kwargs["id"], str)
                    and (len(kwargs["id"].split(".")) == 2)
                ):
                    device_id, ch_id = kwargs["id"].split(".")
                    kwargs["id"] = ch_id
                    return targets[device_id](*args, **kwargs)
                return {k: t(*args, **kwargs) for k, t in targets.items()}

            return func
        else:
            return targets

    def __call__(self, *args, **kwargs) -> "DeviceMapping[str, Any]":
        if (
            ("id" in kwargs)
            and isinstance(kwargs["id"], str)
            and (len(kwargs["id"].split(".")) == 2)
        ):
            device_id, ch_id = kwargs["id"].split(".")
            kwargs["id"] = ch_id
            return self[device_id](*args, **kwargs)
        return DeviceMapping((k, v(*args, **kwargs)) for k, v in self.items())

    def __repr__(self) -> str:
        items = ", ".join([f"{k}={v!r}" for k, v in self.items()])
        return f"DeviceMapping({items})"

    def __str__(self) -> str:
        items = ", ".join([f"{k}={v!s}" for k, v in self.items()])
        return f"DeviceMapping({items})"


class ConfigManager:
    def __set_name__(self, owner, name):
        self.name = "_" + name

    def __get__(self, instance, owner):
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        setattr(instance, self.name, value)


class DeviceBase(ABC):

    _instances: Dict[str, "DeviceBase"] = {}
    Identifier: Optional[str] = None
    Config: ConfigManager()

    Manufacturer: ClassVar[str] = ""
    Model: ClassVar[str]

    @final
    def __new__(
        cls,
        *,
        name: Optional[str] = None,
        model: Optional[Union[str, Dict[str, str]]] = None,
    ):
        if cls is not DeviceBase:
            if name is None:
                raise ValueError("Name for this device must be specified")
            cfg = find_config(name, cls.Identifier)
            key = f"{cls.Manufacturer}.{cls.Model}.{getattr(cfg, cls.Identifier, None)}"
            if key in cls._instances:
                cls._instances[key].__init__ = lambda *args, **kwargs: None
                return cls._instances[key]

            cls._instances[key] = super().__new__(cls)
            cls._instances[key].Config = cfg
            return cls._instances[key]

        impl = get_device_list()

        if isinstance(model, dict):
            name = {k: f"{name}.{k}" for k in model}
            parsed = {
                k: partial(impl[v.lower()], name=name[k]) for k, v in model.items()
            }
            return DeviceMapping(parsed)

        parsed = partial(impl[model.lower()], name=name)
        return parsed

    @final
    def __repr__(self) -> str:
        model = f"model={self.Model!r}"
        manufacturer = f"manufacturer={self.Manufacturer!r}"
        ident = ""
        if self.Identifier is not None:
            ident = f"{self.Identifier}={getattr(self.Config, self.Identifier, None)!r}"
        return f"{self.__class__.__name__}({model}, {manufacturer}, {ident})"

    @final
    def __str__(self) -> str:
        return self.__repr__()

    @abstractmethod
    def finalize(self) -> None:
        ...



from copy import deepcopy
from typing import Sequence, TypeVar

import numpy as np

from neclib.data.interpolate import Interpolator

T = TypeVar("T")
Index = TypeVar("Index", int, float)


class LinearExtrapolate(Interpolator[T]):
    """Extrapolate Python objects.

    Example
    -------
    >>> ext = neclib.data.interpolate.LinearInterp(align_by="x", attrs=["x", "y"])
    >>> x = SimpleNamespace(x=4)
    >>> xs = [SimpleNamespace(x=0, y=0), SimpleNamespace(x=2, y=100)]
    >>> ext(x, xs)
    SimpleNamespace(x=4.0, y=200.0)

    """

    def _extrapolate(self, idx: float, xs: Sequence[T]) -> T:
        ret_x = deepcopy(xs[0])
        if self.attrs is not None:
            for attr in self.attrs:
                try:
                    _xs = [getattr(_x, attr) for _x in xs]
                    slope, intercept = np.polyfit(range(len(xs)), _xs, 1)
                    v = slope * idx + intercept
                    setattr(ret_x, attr, v)
                except Exception:
                    pass
            return ret_x
        else:
            slope, intercept = np.polyfit(range(len(xs)), xs, 1)
            return slope * idx + intercept

    def __call__(self, x: T, xs: Sequence[T]) -> T:
        _x, _xs, xs = self._get_sorted(x, xs)
        slope, intercept = np.polyfit(range(len(_xs)), _xs, 1)
        idx = (_x - intercept) / slope
        return self._extrapolate(idx, xs)
